// AD9850 80m/40m VFO, rotary encoder, LCD 1602
// 26.04.2018, Arduino IDE v1.8.5, LZ2WSG, KN34PC
//---------------------------------------------------------------------------------------------------------
#include <LiquidCrystal.h>
#include "Rotary.h"

#define FQ_UD   8                              // AD9850 FQ_UD
#define W_CLK   9                              // AD9850 W_CLK
#define DATA    10                             // AD9850 DATA
#define F_MCLK  125000000                      // F_MCLK is 125 MHz

#define SW_STEP A0                             // Encoder Increment Button
#define SW_BAND A1                             // BAND switch
#define SW_LOCK A2                             // VFO LOCK switch

#define pulse_high(pin) {digitalWrite(pin, HIGH); digitalWrite(pin, LOW);}

LiquidCrystal lcd(12, 13, 7, 6, 5, 4);         // RS(4), E(6), D4(11), D5(12), D6(13), D7(14)
Rotary r = Rotary(2, 3);

uint32_t rx[2] = {3650000, 7120000};
uint32_t band_l[2] = {3500000, 7000000};
uint32_t band_h[2] = {3800000, 7200000};
uint32_t if_freq = 500000;
uint32_t enc_inc = 100;
int8_t enc_dir = 0;
uint8_t band = 1;
uint8_t band_old = 1;
bool lock = false;
bool lock_old = false;
bool ch_flag = true;
uint8_t bt = 0;
//---------------------------------------------------------------------------------------------------------
void setup() {
  lcd.begin(16, 2);

  pinMode(FQ_UD, OUTPUT);
  pinMode(W_CLK, OUTPUT);
  pinMode(DATA, OUTPUT);

  pinMode(SW_STEP, INPUT_PULLUP);
  pinMode(SW_BAND, INPUT_PULLUP);
  pinMode(SW_LOCK, INPUT_PULLUP);

  attachInterrupt(0, rotary_encoder, CHANGE);
  attachInterrupt(1, rotary_encoder, CHANGE);

  ad9850_init();
  show_inc();
}
//---------------------------------------------------------------------------------------------------------
void loop() {
  if (!lock && ch_flag) {
    rx[band] += enc_dir * enc_inc;
    enc_dir = 0;
    rx[band] = constrain(rx[band], band_l[band], band_h[band]);

    ad9850_set_freq(rx[band] + if_freq);
    show_freq(rx[band]);

    ch_flag = false;
  }

  if (digitalRead(SW_STEP) == LOW) {           // step button
    if (button_state(SW_STEP) > 0) {
      set_inc();
      show_inc();
      do {
      }
      while (digitalRead(SW_STEP) == LOW);
    }
  }

  lock = !digitalRead(SW_LOCK);                // LOCK switch
  if (lock != lock_old) {
    show_lock(lock);
    lock_old = lock;
  }

  band = digitalRead(SW_BAND);                 // BAND switch
  if (band != band_old) {
    ch_flag = true;
    band_old = band;
  }
}
//---------------------------------------------------------------------------------------------------------
void ad9850_init() {
  pulse_high(W_CLK);
  pulse_high(FQ_UD);
}
//---------------------------------------------------------------------------------------------------------
void ad9850_set_freq(double frequency) {
  uint32_t freq = frequency * 4294967296 / F_MCLK;

  noInterrupts();
  shiftOut(DATA, W_CLK, LSBFIRST, freq);
  shiftOut(DATA, W_CLK, LSBFIRST, (freq >> 8));
  shiftOut(DATA, W_CLK, LSBFIRST, (freq >> 16));
  shiftOut(DATA, W_CLK, LSBFIRST, (freq >> 24));
  shiftOut(DATA, W_CLK, LSBFIRST, 0b00000000);
  pulse_high(FQ_UD);
  interrupts();
}
//---------------------------------------------------------------------------------------------------------
void show_freq(uint32_t sf_rx) {
  uint16_t fr = sf_rx / 1000000;

  lcd.setCursor(1, 0);
  if (fr < 10)
    lcd.print(' ');
  lcd.print(fr);
  lcd.print('.');
  fr = (sf_rx % 1000000) / 1000;
  if (fr < 100)
    lcd.print('0');
  if (fr < 10)
    lcd.print('0');
  lcd.print(fr);
  lcd.print('.');
  fr = (sf_rx % 1000) / 10;
  if (fr < 10)
    lcd.print('0');
  lcd.print(fr);
}
//---------------------------------------------------------------------------------------------------------
void show_lock(bool st) {
  lcd.setCursor(0, 0);
  if (st)
    lcd.print('x');
  else
    lcd.print(' ');
}
//---------------------------------------------------------------------------------------------------------
void show_inc() {
  lcd.setCursor(11, 0);
  switch (enc_inc) {
    case 10  : lcd.print("  10"); break;
    case 50  : lcd.print("  50"); break;
    case 100 : lcd.print(" 100"); break;
    case 1000: lcd.print("1000"); break;
  }
}
//---------------------------------------------------------------------------------------------------------
void set_inc() {
  switch (enc_inc) {
    case 10  : enc_inc = 1000; break;
    case 50  : enc_inc = 10; break;
    case 100 : enc_inc = 50; break;
    case 1000: enc_inc = 100; break;
  }
}
//---------------------------------------------------------------------------------------------------------
void rotary_encoder() {
  uint8_t result = r.process();

  if (result) {
    ch_flag = true;
    if (result == DIR_CW)
      enc_dir = 1;
    else
      enc_dir = -1;
  }
}//--------------------------------------------------------------------------------------------------------
uint8_t button_state(uint8_t bt) {               // push button state: 0(very fast), 1(fast), 2(slow)
  uint32_t t_start;
  uint32_t t_bt = 0;
  uint8_t bt_state;

  t_start = millis();
  while (!digitalRead(bt)) {
    t_bt = millis() - t_start;
    if (t_bt > 1000)
      break;
  }
  if (t_bt < 20)                                 // 'nothing' for debounce
    bt_state = 0;
  else if (t_bt < 400)                           // fast
    bt_state = 1;
  else                                           // slow
    bt_state = 2;
  return bt_state;
}
